@@ -20,7 +20,6 @@ use {
         ping_pong::PingCache,
         weighted_shuffle::WeightedShuffle,
     },
-    itertools::Itertools,
     lru::LruCache,
     rand::Rng,
     rayon::{prelude::*, ThreadPool},
@@ -30,12 +29,16 @@ use {
         pubkey::Pubkey,
         signature::{Keypair, Signer},
     },
+    solana_streamer::socket::SocketAddrSpace,
     std::{
         collections::{HashMap, HashSet, VecDeque},
         convert::TryInto,
         iter::repeat_with,
         net::SocketAddr,
-        sync::Mutex,
+        sync::{
+            atomic::{AtomicI64, AtomicUsize, Ordering},
+            Mutex,
+        },
         time::{Duration, Instant},
     },
 };
@@ -221,6 +224,7 @@ impl CrdsGossipPull {
         bloom_size: usize,
         ping_cache: &Mutex<PingCache>,
         pings: &mut Vec<(SocketAddr, Ping)>,
+        socket_addr_space: &SocketAddrSpace,
     ) -> Result<(ContactInfo, Vec<CrdsFilter>), CrdsGossipError> {
         let (weights, peers): (Vec<_>, Vec<_>) = self
             .pull_options(
@@ -230,6 +234,7 @@ impl CrdsGossipPull {
                 now,
                 gossip_validators,
                 stakes,
+                socket_addr_space,
             )
             .into_iter()
             .unzip();
@@ -271,6 +276,7 @@ impl CrdsGossipPull {
         now: u64,
         gossip_validators: Option<&HashSet<Pubkey>>,
         stakes: &HashMap<Pubkey, u64>,
+        socket_addr_space: &SocketAddrSpace,
     ) -> Vec<(u64, &'a ContactInfo)> {
         let mut rng = rand::thread_rng();
         let active_cutoff = now.saturating_sub(PULL_ACTIVE_TIMEOUT_MS);
@@ -290,7 +296,7 @@ impl CrdsGossipPull {
             })
             .filter(|v| {
                 v.id != *self_id
-                    && ContactInfo::is_valid_address(&v.gossip)
+                    && ContactInfo::is_valid_address(&v.gossip, socket_addr_space)
                     && (self_shred_version == 0 || self_shred_version == v.shred_version)
                     && gossip_validators
                         .map_or(true, |gossip_validators| gossip_validators.contains(&v.id))
@@ -334,12 +340,13 @@ impl CrdsGossipPull {
 
     /// Create gossip responses to pull requests
     pub(crate) fn generate_pull_responses(
+        thread_pool: &ThreadPool,
         crds: &Crds,
         requests: &[(CrdsValue, CrdsFilter)],
         output_size_limit: usize, // Limit number of crds values returned.
         now: u64,
     ) -> Vec<Vec<CrdsValue>> {
-        Self::filter_crds_values(crds, requests, output_size_limit, now)
+        Self::filter_crds_values(thread_pool, crds, requests, output_size_limit, now)
     }
 
     // Checks if responses should be inserted and
@@ -472,9 +479,10 @@ impl CrdsGossipPull {
 
     /// filter values that fail the bloom filter up to max_bytes
     fn filter_crds_values(
+        thread_pool: &ThreadPool,
         crds: &Crds,
         filters: &[(CrdsValue, CrdsFilter)],
-        mut output_size_limit: usize, // Limit number of crds values returned.
+        output_size_limit: usize, // Limit number of crds values returned.
         now: u64,
     ) -> Vec<Vec<CrdsValue>> {
         let msg_timeout = CRDS_GOSSIP_PULL_CRDS_TIMEOUT_MS;
@@ -482,49 +490,56 @@ impl CrdsGossipPull {
         //skip filters from callers that are too old
         let caller_wallclock_window =
             now.saturating_sub(msg_timeout)..now.saturating_add(msg_timeout);
-        let mut dropped_requests = 0;
-        let mut total_skipped = 0;
-        let ret: Vec<_> = filters
-            .iter()
-            .map(|(caller, filter)| {
-                if output_size_limit == 0 {
-                    return None;
+        let dropped_requests = AtomicUsize::default();
+        let total_skipped = AtomicUsize::default();
+        let output_size_limit = output_size_limit.try_into().unwrap_or(i64::MAX);
+        let output_size_limit = AtomicI64::new(output_size_limit);
+        let apply_filter = |caller: &CrdsValue, filter: &CrdsFilter| {
+            if output_size_limit.load(Ordering::Relaxed) <= 0 {
+                return Vec::default();
+            }
+            let caller_wallclock = caller.wallclock();
+            if !caller_wallclock_window.contains(&caller_wallclock) {
+                dropped_requests.fetch_add(1, Ordering::Relaxed);
+                return Vec::default();
+            }
+            let caller_pubkey = caller.pubkey();
+            let caller_wallclock = caller_wallclock.checked_add(jitter).unwrap_or(0);
+            let pred = |entry: &&VersionedCrdsValue| {
+                debug_assert!(filter.test_mask(&entry.value_hash));
+                // Skip values that are too new.
+                if entry.value.wallclock() > caller_wallclock {
+                    total_skipped.fetch_add(1, Ordering::Relaxed);
+                    false
+                } else {
+                    !filter.filter_contains(&entry.value_hash)
+                        && (entry.value.pubkey() != caller_pubkey
+                            || entry.value.should_force_push(&caller_pubkey))
                 }
-                let caller_wallclock = caller.wallclock();
-                if !caller_wallclock_window.contains(&caller_wallclock) {
-                    dropped_requests += 1;
-                    return Some(vec![]);
-                }
-                let caller_pubkey = caller.pubkey();
-                let caller_wallclock = caller_wallclock.checked_add(jitter).unwrap_or(0);
-                let pred = |entry: &&VersionedCrdsValue| {
-                    debug_assert!(filter.test_mask(&entry.value_hash));
-                    // Skip values that are too new.
-                    if entry.value.wallclock() > caller_wallclock {
-                        total_skipped += 1;
-                        false
-                    } else {
-                        !filter.filter_contains(&entry.value_hash)
-                            && (entry.value.pubkey() != caller_pubkey
-                                || entry.value.should_force_push(&caller_pubkey))
-                    }
-                };
-                let out: Vec<_> = crds
-                    .filter_bitmask(filter.mask, filter.mask_bits)
-                    .filter(pred)
-                    .map(|entry| entry.value.clone())
-                    .take(output_size_limit)
-                    .collect();
-                output_size_limit -= out.len();
-                Some(out)
-            })
-            .while_some()
-            .collect();
+            };
+            let out: Vec<_> = crds
+                .filter_bitmask(filter.mask, filter.mask_bits)
+                .filter(pred)
+                .map(|entry| entry.value.clone())
+                .take(output_size_limit.load(Ordering::Relaxed).max(0) as usize)
+                .collect();
+            output_size_limit.fetch_sub(out.len() as i64, Ordering::Relaxed);
+            out
+        };
+        let ret: Vec<_> = thread_pool.install(|| {
+            filters
+                .par_iter()
+                .map(|(caller, filter)| apply_filter(caller, filter))
+                .collect()
+        });
         inc_new_counter_info!(
             "gossip_filter_crds_values-dropped_requests",
-            dropped_requests + filters.len() - ret.len()
+            dropped_requests.into_inner()
+        );
+        inc_new_counter_info!(
+            "gossip_filter_crds_values-dropped_values",
+            total_skipped.into_inner()
         );
-        inc_new_counter_info!("gossip_filter_crds_values-dropped_values", total_skipped);
         ret
     }
 
@@ -687,7 +702,15 @@ pub(crate) mod tests {
             stakes.insert(id, i * 100);
         }
         let now = 1024;
-        let mut options = node.pull_options(&crds, &me.label().pubkey(), 0, now, None, &stakes);
+        let mut options = node.pull_options(
+            &crds,
+            &me.label().pubkey(),
+            0,
+            now,
+            None,
+            &stakes,
+            &SocketAddrSpace::Unspecified,
+        );
         assert!(!options.is_empty());
         options.sort_by(|(weight_l, _), (weight_r, _)| weight_r.partial_cmp(weight_l).unwrap());
         // check that the highest stake holder is also the heaviest weighted.
@@ -737,7 +760,15 @@ pub(crate) mod tests {
 
         // shred version 123 should ignore nodes with versions 0 and 456
         let options = node
-            .pull_options(&crds, &me.label().pubkey(), 123, 0, None, &stakes)
+            .pull_options(
+                &crds,
+                &me.label().pubkey(),
+                123,
+                0,
+                None,
+                &stakes,
+                &SocketAddrSpace::Unspecified,
+            )
             .iter()
             .map(|(_, c)| c.id)
             .collect::<Vec<_>>();
@@ -747,7 +778,15 @@ pub(crate) mod tests {
 
         // spy nodes will see all
         let options = node
-            .pull_options(&crds, &spy.label().pubkey(), 0, 0, None, &stakes)
+            .pull_options(
+                &crds,
+                &spy.label().pubkey(),
+                0,
+                0,
+                None,
+                &stakes,
+                &SocketAddrSpace::Unspecified,
+            )
             .iter()
             .map(|(_, c)| c.id)
             .collect::<Vec<_>>();
@@ -787,6 +826,7 @@ pub(crate) mod tests {
             0,
             Some(&gossip_validators),
             &stakes,
+            &SocketAddrSpace::Unspecified,
         );
         assert!(options.is_empty());
 
@@ -799,6 +839,7 @@ pub(crate) mod tests {
             0,
             Some(&gossip_validators),
             &stakes,
+            &SocketAddrSpace::Unspecified,
         );
         assert!(options.is_empty());
 
@@ -811,6 +852,7 @@ pub(crate) mod tests {
             0,
             Some(&gossip_validators),
             &stakes,
+            &SocketAddrSpace::Unspecified,
         );
         assert_eq!(options.len(), 1);
         assert_eq!(options[0].1.id, node_123.pubkey());
@@ -929,6 +971,7 @@ pub(crate) mod tests {
                 PACKET_DATA_SIZE,
                 &ping_cache,
                 &mut pings,
+                &SocketAddrSpace::Unspecified,
             ),
             Err(CrdsGossipError::NoPeers)
         );
@@ -946,6 +989,7 @@ pub(crate) mod tests {
                 PACKET_DATA_SIZE,
                 &ping_cache,
                 &mut pings,
+                &SocketAddrSpace::Unspecified,
             ),
             Err(CrdsGossipError::NoPeers)
         );
@@ -968,6 +1012,7 @@ pub(crate) mod tests {
             PACKET_DATA_SIZE,
             &ping_cache,
             &mut pings,
+            &SocketAddrSpace::Unspecified,
         );
         let (peer, _) = req.unwrap();
         assert_eq!(peer, *new.contact_info().unwrap());
@@ -987,6 +1032,7 @@ pub(crate) mod tests {
             PACKET_DATA_SIZE,
             &ping_cache,
             &mut pings,
+            &SocketAddrSpace::Unspecified,
         );
         // Even though the offline node should have higher weight, we shouldn't request from it
         // until we receive a ping.
@@ -1041,6 +1087,7 @@ pub(crate) mod tests {
                     PACKET_DATA_SIZE, // bloom_size
                     &ping_cache,
                     &mut pings,
+                    &SocketAddrSpace::Unspecified,
                 )
                 .unwrap();
             peer
@@ -1118,16 +1165,18 @@ pub(crate) mod tests {
             PACKET_DATA_SIZE,
             &Mutex::new(ping_cache),
             &mut pings,
+            &SocketAddrSpace::Unspecified,
         );
 
         let mut dest_crds = Crds::default();
         let (_, filters) = req.unwrap();
         let mut filters: Vec<_> = filters.into_iter().map(|f| (caller.clone(), f)).collect();
         let rsp = CrdsGossipPull::generate_pull_responses(
+            &thread_pool,
             &dest_crds,
             &filters,
-            /*output_size_limit=*/ usize::MAX,
-            0,
+            usize::MAX, // output_size_limit
+            0,          // now
         );
 
         assert_eq!(rsp[0].len(), 0);
@@ -1142,10 +1191,11 @@ pub(crate) mod tests {
 
         //should skip new value since caller is to old
         let rsp = CrdsGossipPull::generate_pull_responses(
+            &thread_pool,
             &dest_crds,
             &filters,
-            /*output_size_limit=*/ usize::MAX,
-            CRDS_GOSSIP_PULL_MSG_TIMEOUT_MS,
+            usize::MAX,                      // output_size_limit
+            CRDS_GOSSIP_PULL_MSG_TIMEOUT_MS, // now
         );
         assert_eq!(rsp[0].len(), 0);
         assert_eq!(filters.len(), MIN_NUM_BLOOM_FILTERS);
@@ -1160,9 +1210,10 @@ pub(crate) mod tests {
                 .collect::<Vec<_>>()
         });
         let rsp = CrdsGossipPull::generate_pull_responses(
+            &thread_pool,
             &dest_crds,
             &filters,
-            /*output_size_limit=*/ usize::MAX,
+            usize::MAX, // output_size_limit
             CRDS_GOSSIP_PULL_MSG_TIMEOUT_MS,
         );
         assert_eq!(rsp.len(), 2 * MIN_NUM_BLOOM_FILTERS);
@@ -1205,16 +1256,18 @@ pub(crate) mod tests {
             PACKET_DATA_SIZE,
             &Mutex::new(ping_cache),
             &mut pings,
+            &SocketAddrSpace::Unspecified,
         );
 
         let mut dest_crds = Crds::default();
         let (_, filters) = req.unwrap();
         let filters: Vec<_> = filters.into_iter().map(|f| (caller.clone(), f)).collect();
         let rsp = CrdsGossipPull::generate_pull_responses(
+            &thread_pool,
             &dest_crds,
             &filters,
-            /*output_size_limit=*/ usize::MAX,
-            0,
+            usize::MAX, // output_size_limit
+            0,          // now
         );
         CrdsGossipPull::process_pull_requests(
             &mut dest_crds,
@@ -1278,14 +1331,16 @@ pub(crate) mod tests {
                 PACKET_DATA_SIZE,
                 &ping_cache,
                 &mut pings,
+                &SocketAddrSpace::Unspecified,
             );
             let (_, filters) = req.unwrap();
             let filters: Vec<_> = filters.into_iter().map(|f| (caller.clone(), f)).collect();
             let rsp = CrdsGossipPull::generate_pull_responses(
+                &thread_pool,
                 &dest_crds,
                 &filters,
-                /*output_size_limit=*/ usize::MAX,
-                0,
+                usize::MAX, // output_size_limit
+                0,          // now
             );
             CrdsGossipPull::process_pull_requests(
                 &mut dest_crds,
